#include "flatplate.h"
double dot2(double *a, double *b, double *t, int Nt)
{
	double sum = 0;
	int kk;
	for(kk=1;kk<Nt-1;kk++) sum += a[kk]*b[kk]*(t[kk+1]-t[kk]);
	// for(kk=1;kk<Nt-1;kk++) sum += a[kk]*b[kk];
	return sum;
}

double grid(double kk, int Nt, double L1, double L2, char type)
{
	double s = kk*1.0/(Nt-1);
	double Lx = L2-L1;
	// // Linear
	if (type==1) return L1 + s*Lx;

	// // Quadratic
	if (type==2) return L1 + Lx*s*s*(3-2*s);

	// // Cubic
	if (type==3) return L1 + Lx*s*s*s*(10-15*s+6*s*s);
	
	// // Quartic
	if (type==4) return L1 + Lx*pow(s,4)*(35 - 84*s + 70*s*s - 20*s*s*s);
	
	// // Exponential
	if (type==-1)
	{
	    const double lam = 5;
	    const double fac = Lx/(1+exp(-lam)/lam - 1.0/lam);
	   return L1 + fac*(s + exp(-lam*s)/lam - 1.0/lam);
	}
	else return L1 + s*Lx;
}

struct State state_new(int Nx, int Ny, double Lx1, double Lx2, double Ly)
{
	int ii, jj;
	struct State S;
	double *x, *y;
	S.Nx = Nx;
	S.Ny = Ny;
	S.Lx1 = Lx1;
	S.Lx2 = Lx2;
	S.Ly = Ly;
	S.x = (double*) malloc(Nx*sizeof(double));
	S.y = (double*) malloc(Ny*sizeof(double));
	S.xh = (double*) malloc((Nx-1)*sizeof(double));
	S.yh = (double*) malloc((Ny-1)*sizeof(double));
	S.u = (double*) malloc(Nx*(Ny-1)*sizeof(double));
	S.v = (double*) malloc((Nx-1)*Ny*sizeof(double));
	S.unew = (double*) malloc(Nx*(Ny-1)*sizeof(double));

	S.a    = (double*) malloc(Nx*(Ny-1)*sizeof(double));
	S.anew = (double*) malloc(Nx*(Ny-1)*sizeof(double));
	S.b    = (double*) malloc((Nx-1)*(Ny-1)*sizeof(double));

	S.uy = (double*) malloc(Nx*sizeof(double));
	S.ay = (double*) malloc(Nx*sizeof(double));

	x = S.x;
	y = S.y;
	for(ii=0;ii<Nx;ii++) x[ii] = grid(ii, Nx, Lx1, Lx2, 1);
	/*
	{
		int i1, i2;
		i1 = (Nx-1)/4+1; // 11
		i2 = (Nx-1)*3/4 + 1; // 31
		for(ii=0;  ii<i1; ii++) x[i1-ii-1] = grid(ii, 2*i1-1, 0, 2*Lx1, 2);
		for(ii=i1; ii<i2; ii++) x[ii] = grid(ii-i1+1, (i2-i1)+1, 0, 1, 2);
		for(ii=i2; ii<Nx; ii++) x[ii] = grid(ii-i2+1, Nx-i2+1,   1, Lx2, 2);
	}
	*/

	// for(jj=0;jj<Ny;jj++) y[jj] = grid(jj, Ny, 0, Ly, 1);
	for(jj=0;jj<Ny;jj++) y[jj] = grid(jj, Ny, 0, Ly, -1);

	for(ii=0;ii<Nx-1;ii++) S.xh[ii] = 0.5*(x[ii]+x[ii+1]);
	for(jj=0;jj<Ny-1;jj++) S.yh[jj] = 0.5*(y[jj]+y[jj+1]);

	for(ii=0;ii<Nx  ;ii++) for(jj=0;jj<Ny-1;jj++) elu(S.u,   jj, ii) = 0;
	for(ii=0;ii<Nx-1;ii++) for(jj=0;jj<Ny  ;jj++) elv(S.v,   jj, ii) = 0;

	return S;
}

double trapz(double *time, double *f, int Nt)
{
	double sum = 0;
	int kk;
	for(kk=1;kk<Nt;kk++) sum += 0.5*(f[kk-1]+f[kk])*(time[kk]-time[kk-1]);
	return sum;
}

void state_writefsnap(hid_t fileid, int nn, struct State S)
{
	hid_t       usetid, vsetid, dataspace_id;
	hid_t       uslabid, vslabid, memspace_id;
	hsize_t     dims1[1], dims2[2], dims3[3], dims3b[3];
	hsize_t     start[3], stride[3], count[3], block[3];
	herr_t      status;

	usetid = H5Dopen(fileid, "/u", H5P_DEFAULT);
	dims3b[0]  = (S.Ny-1)*S.Nx; 

	start[0]  = nn; start[1]  = 0;       start[2]  = 0;
	stride[0] = 1;  stride[1] = 1;       stride[2] = 1;
	count[0]  = 1;  count[2]  = S.Ny-1;  count[1]  = S.Nx;
	block[0]  = 1;  block[1]  = 1;       block[2]  = 1;

	memspace_id  = H5Screate_simple (1, dims3b, NULL); 
	uslabid      = H5Dget_space(usetid);
	status       = H5Sselect_hyperslab(uslabid, H5S_SELECT_SET, start, NULL, count, NULL);
	status       = H5Dwrite(usetid, H5T_NATIVE_DOUBLE, memspace_id, uslabid, H5P_DEFAULT, S.u);
	H5Dclose(usetid);

	vsetid = H5Dopen(fileid, "/v", H5P_DEFAULT);
	dims3b[0]  = (S.Nx-1)*S.Ny; 

	start[0]  = nn; start[1]  = 0;       start[2]  = 0;
	stride[0] = 1;  stride[1] = 1;       stride[2] = 1;
	count[0]  = 1;  count[2]  = S.Ny;    count[1]  = S.Nx-1;
	block[0]  = 1;  block[1]  = 1;       block[2]  = 1;

	memspace_id  = H5Screate_simple (1, dims3b, NULL); 
	vslabid      = H5Dget_space(vsetid);
	status       = H5Sselect_hyperslab(vslabid, H5S_SELECT_SET, start, NULL, count, NULL);
	status       = H5Dwrite(vsetid, H5T_NATIVE_DOUBLE, memspace_id, vslabid, H5P_DEFAULT, S.v);
	H5Dclose(vsetid);

	usetid = H5Dopen(fileid, "/uys", H5P_DEFAULT);
	dims3b[0]  = S.Nx; 

	start[0]  = nn; start[1]  = 0;   
	stride[0] = 1;  stride[1] = 1;   
	count[0]  = 1;  count[1]  = S.Nx;
	block[0]  = 1;  block[1]  = 1;   

	memspace_id  = H5Screate_simple (1, dims3b, NULL); 
	uslabid      = H5Dget_space(usetid);
	status       = H5Sselect_hyperslab(uslabid, H5S_SELECT_SET, start, NULL, count, NULL);
	status       = H5Dwrite(usetid, H5T_NATIVE_DOUBLE, memspace_id, uslabid, H5P_DEFAULT, S.uy);
	H5Dclose(usetid);
	return;
}
void state_readfsnap(hid_t fileid, int nn, struct State S)
{
	hid_t       usetid, vsetid, dataspace_id;
	hid_t       uslabid, vslabid, memspace_id;
	hsize_t     dims1[1], dims2[2], dims3[3], dims3b[3];
	hsize_t     start[3], stride[3], count[3], block[3];
	herr_t      status;

	usetid = H5Dopen(fileid, "/u", H5P_DEFAULT);
	dims3b[0]  = (S.Ny-1)*S.Nx; 

	start[0]  = nn; start[1]  = 0;       start[2]  = 0;
	stride[0] = 1;  stride[1] = 1;       stride[2] = 1;
	count[0]  = 1;  count[2]  = S.Ny-1;  count[1]  = S.Nx;
	block[0]  = 1;  block[1]  = 1;       block[2]  = 1;

	memspace_id  = H5Screate_simple (1, dims3b, NULL); 
	uslabid      = H5Dget_space(usetid);
	status       = H5Sselect_hyperslab(uslabid, H5S_SELECT_SET, start, NULL, count, NULL);
	status       = H5Dread(usetid, H5T_NATIVE_DOUBLE, memspace_id, uslabid, H5P_DEFAULT, S.u);
	H5Dclose(usetid);

	vsetid = H5Dopen(fileid, "/v", H5P_DEFAULT);
	dims3b[0]  = (S.Nx-1)*S.Ny; 

	start[0]  = nn; start[1]  = 0;       start[2]  = 0;
	stride[0] = 1;  stride[1] = 1;       stride[2] = 1;
	count[0]  = 1;  count[2]  = S.Ny;    count[1]  = S.Nx-1;
	block[0]  = 1;  block[1]  = 1;       block[2]  = 1;

	memspace_id  = H5Screate_simple (1, dims3b, NULL); 
	vslabid      = H5Dget_space(vsetid);
	status       = H5Sselect_hyperslab(vslabid, H5S_SELECT_SET, start, NULL, count, NULL);
	status       = H5Dread(vsetid, H5T_NATIVE_DOUBLE, memspace_id, vslabid, H5P_DEFAULT, S.v);
	H5Dclose(vsetid);
	return;
}

void state_writebsnap(hid_t fileid, int nn, struct State S)
{
	hid_t       asetid, bsetid, dataspace_id;
	hid_t       aslabid, bslabid, memspace_id;
	hsize_t     dims1[1], dims2[2], dims3[3], dims3b[3];
	hsize_t     start[3], stride[3], count[3], block[3];
	herr_t      status;

	asetid = H5Dopen(fileid, "/a", H5P_DEFAULT);
	dims3b[0]  = (S.Ny-1)*S.Nx; 

	start[0]  = nn; start[1]  = 0;       start[2]  = 0;
	stride[0] = 1;  stride[1] = 1;       stride[2] = 1;
	count[0]  = 1;  count[2]  = S.Ny-1;  count[1]  = S.Nx;
	block[0]  = 1;  block[1]  = 1;       block[2]  = 1;

	memspace_id  = H5Screate_simple (1, dims3b, NULL); 
	aslabid      = H5Dget_space(asetid);
	status       = H5Sselect_hyperslab(aslabid, H5S_SELECT_SET, start, NULL, count, NULL);
	status       = H5Dwrite(asetid, H5T_NATIVE_DOUBLE, memspace_id, aslabid, H5P_DEFAULT, S.a);
	H5Dclose(asetid);

	bsetid = H5Dopen(fileid, "/b", H5P_DEFAULT);
	dims3b[0]  = (S.Ny-1)*(S.Nx-1); 

	start[0]  = nn; start[1]  = 0;       start[2]  = 0;
	stride[0] = 1;  stride[1] = 1;       stride[2] = 1;
	count[0]  = 1;  count[2]  = S.Ny-1;  count[1]  = S.Nx-1;
	block[0]  = 1;  block[1]  = 1;       block[2]  = 1;

	memspace_id  = H5Screate_simple (1, dims3b, NULL); 
	bslabid      = H5Dget_space(bsetid);
	status       = H5Sselect_hyperslab(bslabid, H5S_SELECT_SET, start, NULL, count, NULL);
	status       = H5Dwrite(bsetid, H5T_NATIVE_DOUBLE, memspace_id, bslabid, H5P_DEFAULT, S.b);
	H5Dclose(bsetid);

	asetid = H5Dopen(fileid, "/ays", H5P_DEFAULT);
	dims3b[0]  = S.Nx; 

	start[0]  = nn; start[1]  = 0;   
	stride[0] = 1;  stride[1] = 1;   
	count[0]  = 1;  count[1]  = S.Nx;
	block[0]  = 1;  block[1]  = 1;   

	memspace_id  = H5Screate_simple (1, dims3b, NULL); 
	aslabid      = H5Dget_space(asetid);
	status       = H5Sselect_hyperslab(aslabid, H5S_SELECT_SET, start, NULL, count, NULL);
	status       = H5Dwrite(asetid, H5T_NATIVE_DOUBLE, memspace_id, aslabid, H5P_DEFAULT, S.ay);
	H5Dclose(asetid);

	return;
}

void state_readbsnap(hid_t fileid, int nn, struct State S)
{
	hid_t       asetid, bsetid, dataspace_id;
	hid_t       aslabid, bslabid, memspace_id;
	hsize_t     dims1[1], dims2[2], dims3[3], dims3b[3];
	hsize_t     start[3], stride[3], count[3], block[3];
	herr_t      status;

	asetid = H5Dopen(fileid, "/a", H5P_DEFAULT);
	dims3b[0]  = (S.Ny-1)*S.Nx; 

	start[0]  = nn; start[1]  = 0;       start[2]  = 0;
	stride[0] = 1;  stride[1] = 1;       stride[2] = 1;
	count[0]  = 1;  count[2]  = S.Ny-1;  count[1]  = S.Nx;
	block[0]  = 1;  block[1]  = 1;       block[2]  = 1;

	memspace_id  = H5Screate_simple (1, dims3b, NULL); 
	aslabid      = H5Dget_space(asetid);
	status       = H5Sselect_hyperslab(aslabid, H5S_SELECT_SET, start, NULL, count, NULL);
	status       = H5Dread(asetid, H5T_NATIVE_DOUBLE, memspace_id, aslabid, H5P_DEFAULT, S.a);
	H5Dclose(asetid);

	bsetid = H5Dopen(fileid, "/b", H5P_DEFAULT);
	dims3b[0]  = (S.Ny-1)*(S.Nx-1); 

	start[0]  = nn; start[1]  = 0;       start[2]  = 0;
	stride[0] = 1;  stride[1] = 1;       stride[2] = 1;
	count[0]  = 1;  count[2]  = S.Ny-1;  count[1]  = S.Nx-1;
	block[0]  = 1;  block[1]  = 1;       block[2]  = 1;

	memspace_id  = H5Screate_simple (1, dims3b, NULL); 
	bslabid      = H5Dget_space(bsetid);
	status       = H5Sselect_hyperslab(bslabid, H5S_SELECT_SET, start, NULL, count, NULL);
	status       = H5Dread(bsetid, H5T_NATIVE_DOUBLE, memspace_id, bslabid, H5P_DEFAULT, S.b);
	H5Dclose(bsetid);

	return;
}

hid_t inith5(char *fname, struct State S, int Nt, double *t, double *f)
{
	hid_t       file_id, usetid, vsetid, xsetid, ysetid, tsetid, fsetid, dataspace_id;
	hid_t       uslabid, vslabid, memspace_id;
	hsize_t     dims1[1], dims2[2], dims3[3], dims3b[3];
	hsize_t     start[3], stride[3], count[3], block[3];
	herr_t      status;

	file_id = H5Fcreate(fname, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);

	dims1[0] = S.Nx;
	dataspace_id = H5Screate_simple(1, dims1, NULL);
	xsetid = H5Dcreate2(file_id, "/x", H5T_NATIVE_DOUBLE, dataspace_id, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
	status = H5Dwrite(xsetid, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, S.x);

	dims1[0] = S.Ny;
	dataspace_id = H5Screate_simple(1, dims1, NULL);
	ysetid = H5Dcreate2(file_id, "/y", H5T_NATIVE_DOUBLE, dataspace_id, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
	status = H5Dwrite(ysetid, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, S.y);

	dims1[0] = Nt;
	dataspace_id = H5Screate_simple(1, dims1, NULL);
	tsetid = H5Dcreate2(file_id, "/t", H5T_NATIVE_DOUBLE, dataspace_id, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
	status = H5Dwrite(tsetid, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, t);

	fsetid = H5Dcreate2(file_id, "/f", H5T_NATIVE_DOUBLE, dataspace_id, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
	status = H5Dwrite(fsetid, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, f);

	dims3[0] = Nt;
	dims3[1] = S.Nx;
	dims3[2] = S.Ny-1;
	dataspace_id = H5Screate_simple(3, dims3, NULL);
	usetid = H5Dcreate2(file_id, "/u", H5T_NATIVE_DOUBLE, dataspace_id, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);

	dims3[0] = Nt;   
	dims3[1] = S.Nx-1; 
	dims3[2] = S.Ny;
	dataspace_id = H5Screate_simple(3, dims3, NULL);
	vsetid       = H5Dcreate2(file_id, "/v", H5T_NATIVE_DOUBLE, dataspace_id, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);

	dims3[0] = Nt;
	dims3[1] = S.Nx;
	dims3[2] = S.Ny-1;
	dataspace_id = H5Screate_simple(3, dims3, NULL);
	usetid = H5Dcreate2(file_id, "/a", H5T_NATIVE_DOUBLE, dataspace_id, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);

	dims3[0] = Nt;
	dims3[1] = S.Nx-1;
	dims3[2] = S.Ny-1;
	dataspace_id = H5Screate_simple(3, dims3, NULL);
	usetid = H5Dcreate2(file_id, "/b", H5T_NATIVE_DOUBLE, dataspace_id, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);

	return file_id; 
}

void vfromu(struct State S)
{
	double *u = S.u;
	double *v = S.v;
	double *x = S.x;
	double *y = S.y;
	int Ny = S.Ny;
	int Nx = S.Nx;

	int ii;
	#pragma omp parallel for
	for(ii=0;ii<Nx-1;ii++)
	{
		int jj;
		elv(v, 0, ii) = 0;
		for(jj=1;jj<Ny;jj++) elv(v,jj,ii) = elv(v, jj-1,ii) + (y[jj] - y[jj-1])*(elu(u, jj-1,ii) - elu(u,jj-1,ii+1))/(x[ii+1]-x[ii]);
	}
	return;
}

void bfroma(struct State S, double fkk)
{
	double *u = S.u;
	double *v = S.v;
	double *a = S.a;
	double *b = S.b;
	int Ny = S.Ny;
	int Nx = S.Nx;

	int ii;
	// printf("In bfroma ... \n");
	#pragma omp parallel for
	for(ii=0;ii<Nx-1;ii++)
	{
		int jj;
		double tmp = 0;
		elu(S.b, Ny-2, ii) = 0;
		for(jj=Ny-2;jj>0; jj--)
		{
			// Here we relate b[ii, jj-1] to b[ii,jj]
			// In grid notation b[ii+1/2, jj-1/2] to b[ii+1/2, jj+1/2]
			// We must have u[kk]=S.unew at the same timestep as a[kk]=S.a
			// and v[kk-1]=S.v to be one earlier time step.
			// 
			double dtmp = 0;
			double vtest, deltau, dx;
			double vsw, vnw, vse, vne;
			double um, up, uc, ac;
			double xp, xm;
			// Top right
			vsw = elv(S.v, jj, ii);
			vnw = elv(S.v, jj+1,ii  );
			vse = (ii<Nx-1)? elv(S.v, jj,ii+1): vsw;
			vne = (ii<Nx-1)? elv(S.v, jj+1,ii+1) : vnw;
			vtest = 0.25*(vsw + vnw + vse + vne);

			ac = elu(S.a,    jj,   ii+1);
			uc = elu(S.unew, jj,   ii+1);
			um = elu(S.unew, jj-1, ii+1);
			up = (jj<Ny-2)? elu(S.unew, jj+1, ii+1): -elu(S.unew, jj, ii+1);
			deltau = (vtest>0) ? (uc - um) : (up - uc);
			xp = (ii<Nx-2) ? S.xh[ii+1] : (2*S.Lx2 - S.xh[ii]);
			xm = S.xh[ii];
			dx = xp - xm;
			dtmp += deltau * ac * dx;

			// Top left
			vsw = (ii>0) ? elv(S.v, jj  , ii-1) : 0;
			vnw = (ii>0) ? elv(S.v, jj+1, ii-1) : 0;
			vse = elv(S.v, jj,   ii);
			vne = elv(S.v, jj+1, ii);
			vtest = 0.25*(vsw + vnw + vse + vne);

			ac = elu(S.a,    jj,   ii);
			uc = elu(S.unew, jj,   ii);
			um = elu(S.unew, jj-1, ii);
			up = (jj<Ny-2)? elu(S.unew, jj+1, ii): -elu(S.unew, jj, ii);
			deltau = (vtest>0) ? (uc - um) : (up - uc);
			xp = S.xh[ii];
			xm = (ii>0) ? S.xh[ii-1] : (2*S.Lx1 - S.xh[ii]);
			dx = xp - xm;
			dtmp += deltau * ac * dx;

			// Bottom right
			vsw = elv(S.v, jj-1, ii);
			vnw = elv(S.v, jj,ii  );
			vse = (ii<Nx-1)? elv(S.v, jj-1, ii+1) : vsw;
			vne = (ii<Nx-1)? elv(S.v, jj,   ii+1) : vnw;
			vtest = 0.25*(vsw + vnw + vse + vne);

			ac = elu(S.a,    jj-1,   ii+1);
			uc = elu(S.unew, jj-1,   ii+1);
			um = (jj>1) ? elu(S.unew, jj-2, ii+1) : 2*fkk-elu(S.unew, 0, ii+1);
			up = elu(S.unew, jj, ii+1);
			deltau = (vtest>0) ? (uc - um) : (up - uc);
			xp = (ii<Nx-2) ? S.xh[ii+1] : (2*S.Lx2 - S.xh[ii]);
			xm = S.xh[ii];
			dx = xp - xm;
			dtmp += deltau * ac * dx;

			// Bottom left
			vsw = (ii>0) ? elv(S.v, jj-1, ii-1) : 0;
			vnw = (ii>0) ? elv(S.v, jj,   ii-1) : 0;
			vse = elv(S.v, jj-1, ii);
			vne = elv(S.v, jj,   ii);
			vtest = 0.25*(vsw + vnw + vse + vne);

			ac = elu(S.a,    jj-1,   ii);
			uc = elu(S.unew, jj-1,   ii);
			um = (jj>1)? elu(S.unew, jj-2, ii): 2*fkk - elu(S.unew, 0, ii);
			up = elu(S.unew, jj, ii);
			deltau = (vtest>0) ? (uc - um) : (up - uc);
			xp = S.xh[ii];
			xm = (ii>0) ? S.xh[ii-1] : (2*S.Lx1 - S.xh[ii]);
			dx = xp - xm;
			dtmp += deltau * ac * dx;

			// Accumulation done, now scale and update the next index of b
			dtmp *= 0.25/(S.x[ii+1] - S.x[ii]);
			tmp -= dtmp;
			elu(b, jj-1, ii) = tmp;
			// if(fabs(tmp)>1e-10) printf("b[%d] = %le\n", jj, tmp);
		}
	}
}

void fwdstep(struct State S, double dt, double *diags, double fkk, double eps, int ncpu)
{
	int nc;
	int Nx = S.Nx, Ny = S.Ny;
	int id[ncpu+1];
	id[0] = 0;
	id[ncpu] = Nx;
	for(nc=1;nc<ncpu;nc++) id[nc] = round(nc*Nx/ncpu);
	#pragma omp parallel for
	for(nc=0;nc<ncpu; nc++)
	{
		double Lx = S.Lx2, Ly = S.Ly;
		double *d, *dl, *du;
		int ii, jj;
		dl = diags + nc*3*(Ny-1);
		d  = dl + (Ny-1);
		du = d  + (Ny-1);
		for(ii=id[nc];ii<id[nc+1];ii++)
		{
			double *r = &(elu(S.unew, 0, ii));
			double dx = ((ii<Nx-1)? S.xh[ii] : (2*S.Lx2 - S.xh[ii-1])) - ((ii>0)? S.xh[ii-1]:-S.xh[0]);
			for(jj=0;jj<Ny-1;jj++)
			{
				double dyp = (jj<(Ny-2)) ? (S.yh[jj+1] - S.yh[jj  ]) : 2*(Ly-S.yh[jj]);
				double dym = (jj>0)      ? (S.yh[jj  ] - S.yh[jj-1]) : 2*S.yh[jj];
				// double fac1 = 2/(dyp+dym);
				double fac1 = 1.0/(S.y[jj+1]-S.y[jj]);
				double rhs;
				double umid, ux;
				// double dx;
			        double duu;
				double vmidplus, vmidminus, vmid;

				dl[jj] = -dt*fac1/dym;
				d [jj] = 1 + dt*fac1*(1/dym + 1/dyp);
				du[jj] = -dt*fac1/dyp;
				rhs = elu(S.u, jj, ii);

				umid = elu(S.u, jj, ii);
				if (umid + fkk>0)
				{
					duu = umid - ( (ii>0) ? elu(S.u, jj, ii-1) : 0 ) ;
					// dx = (ii>0) ? (S.x[ii]   - S.x[ii-1]) : (S.x[ii+1] - S.x[ii  ]);
				}
				else
				{
					duu = ( (ii<Nx-1) ? (elu(S.u, jj, ii+1) - umid) : 0) ;
					// dx = (ii<Nx-1) ? (S.x[ii+1] - S.x[ii]  ) : (S.x[ii  ] - S.x[ii-1]);
				}
				ux = duu/dx;
				rhs -= dt*eps*(umid+fkk)*ux;

				vmidplus  = (ii<Nx-1) ? (0.5*(elv(S.v, jj, ii  ) + elv(S.v, jj+1, ii  ))) : 0; 
				vmidminus = (ii>0)    ? (0.5*(elv(S.v, jj, ii-1) + elv(S.v, jj+1, ii-1))) : 0;
				vmid = 0.5*(vmidplus + vmidminus);
				if (vmid>0) { d[jj] += eps*dt*vmid/dym; dl[jj] -= eps*dt*vmid/dym; };
				if (vmid<0) { d[jj] -= eps*dt*vmid/dyp; du[jj] += eps*dt*vmid/dyp; };
				
				r [jj] = rhs;
			}
			if(S.x[ii]<=1.0+1e-12 && S.x[ii]>=0-1e-12) // No slip for x<=1
			{
				d[0] -= dl[0]; 
				r[0] += 2*fkk*dl[0]; 
			}
			else d[0] += dl[0]; // Otherwise symmetry for x>1
			dl[0] = 0;

			d[Ny-2] -= du[Ny-2]; 
			du[Ny-2] = 0;

			{
				int N = Ny-1, nrhs = 1, info;
				dgtsv_(&N, &nrhs, dl+1, d, du, r, &N, &info);
			}
		}
	}
	return;
}

void bwdstep(struct State S, double dt, double *diags, double fkk, double eps, int ncpu)
{
	int nc;
	int Nx = S.Nx, Ny = S.Ny;
	int id[ncpu+1];
	id[0] = 0;
	id[ncpu] = Nx;
	for(nc=1;nc<ncpu;nc++) id[nc] = round(nc*Nx/ncpu);
	
	#pragma omp parallel for
	for(nc=0;nc<ncpu; nc++)
	{
		double Lx = S.Lx2, Ly = S.Ly;
		double *d, *dl, *du;
		int ii, jj;
		dl = diags + nc*3*(Ny-1);
		d  = dl + (Ny-1);
		du = d  + (Ny-1);
		for(ii=id[nc];ii<id[nc+1];ii++)
		{
			double *r = &(elu(S.anew, 0, ii));
			for(jj=0;jj<Ny-1;jj++)
			{
				double dyp = (jj<(Ny-2)) ? (S.yh[jj+1] - S.yh[jj  ]) : 2*(Ly-S.yh[jj]);
				double dym = (jj>0)      ? (S.yh[jj  ] - S.yh[jj-1]) : 2*S.yh[jj];
				// double fac1 = 2/(dyp+dym);
				double fac1 = 1.0/(S.y[jj+1]-S.y[jj]);
				double rhs;
				double up, um, uc, term;
				double ap, am, ac;
				double xp, xm, dx;
				double vpnw, vpne, vpsw, vpse;
				double vcnw, vcne, vcsw, vcse;
				double vmnw, vmne, vmsw, vmse;
				double vp, vc, vm;
				int jidx;
				
				dl[jj] = -dt*fac1/dym;
				d [jj] = 1 + dt*fac1*(1/dym + 1/dyp);
				du[jj] = -dt*fac1/dyp;
				rhs = elu(S.a, jj, ii);

				uc = elu(S.unew, jj, ii);
				up = (ii<Nx-1)?elu(S.unew, jj, ii+1):elu(S.unew, jj, ii);
				um = (ii>0   )?elu(S.unew, jj, ii-1):0;

				xp = (ii<Nx-1)?S.xh[ii]:2*S.Lx2-S.xh[ii];
				xm = (ii>0   )?S.xh[ii-1]:2*S.Lx1-S.xh[ii];
				dx = xp - xm;
				
				ac = elu(S.a, jj, ii);
				ap = (ii<Nx-1)?elu(S.a, jj, ii+1):0;
				am = (ii>0   )?elu(S.a, jj, ii-1):elu(S.a,jj,ii);
				term = 0;
				term -= (uc+fkk)*((uc+fkk>0)?ac:-ac);
				term += (up+fkk)*((up+fkk>0)?ap:0);
				term -= (um+fkk)*((um+fkk>0)?0:am);

				term -= ac*((uc+fkk>0)?(uc-um):(up-uc));
				rhs += dt*eps*term/dx;

				// Next line is the b_x term
				// rhs += dt*eps* ((ii>0)? ((elu(S.b, jj, ii) - elu(S.b, jj, ii-1))/(S.x[ii] - S.x[ii-1])) : 0);
				rhs += dt*eps* ((ii>0)? ((elu(S.b, jj, ii) - elu(S.b, jj, ii-1))/dx) : 0);
				
				// The implicit discretization of (va)_y term
				// Here use v from time step kk, so S.v

				// We will need v at 12 points
				vcne = (ii<Nx-1)?elv(S.v, jj+1, ii):elv(S.v,jj+1,ii-1);
				vcse = (ii<Nx-1)?elv(S.v, jj,   ii):elv(S.v,jj,  ii-1);
				vcnw = (ii>0)?elv(S.v, jj+1, ii-1):0;
				vcsw = (ii>0)?elv(S.v, jj,   ii-1):0;

				// We have double conditional on ii and jj
				jidx = (jj<Ny-2)?jj+2:jj+1;
				vpne = (ii<Nx-1)?elv(S.v, jidx, ii):elv(S.v,jidx,ii-1);
				vpnw = (ii>0   )?elv(S.v, jidx, ii-1):0;
				vpse = vcne; // (ii<Nx-1)?elv(S.v, jj+1, ii):elv(S.v,jj+1,ii-1);
				vpsw = vcnw; // (ii>0   )?elv(S.v, jj+1, ii-1):0;

				jidx = (jj>0)?jj-1:jj;
				vmne = vcse;
				vmnw = vcsw;
				vmse = (ii<Nx-1)?elv(S.v, jidx, ii  ):elv(S.v,jidx,ii-1);
				vmsw = (ii>0   )?elv(S.v, jidx, ii-1):0;

				vc = 0.25*(vcne+vcnw+vcse+vcsw)*fac1;
				vp = 0.25*(vpne+vpnw+vpse+vpsw)*fac1;
				vm = 0.25*(vmne+vmnw+vmse+vmsw)*fac1;

				d[jj]  += eps*dt* (vc>0?vc:-vc );
				du[jj] += eps*dt* (vp>0?-vp:0);
				dl[jj] += eps*dt* (vm>0?0:vm);
				r[jj] = rhs;
			}
			if(S.x[ii]<=1.0+1e-12 & S.x[ii]>=0-1e-12) // No slip for 0 <= x <= 1
			{
				d[0] -= dl[0]; 
				r[0] -= 2*fkk*dl[0]; 
			}
			else d[0] += dl[0]; // Otherwise symmetry for x>1
			dl[0] = 0;

			d[Ny-2] -= du[Ny-2]; 
			du[Ny-2] = 0;

			{
				int N = Ny-1, nrhs = 1, info;
				dgtsv_(&N, &nrhs, dl+1, d, du, r, &N, &info);
			}
		}

	}
}

int read_attr_int(hid_t fileid, char *attr)
{
	hid_t nxid;
	int Nx=-1;
	nxid = H5Aopen(fileid, attr, H5P_DEFAULT);
	H5Aread(nxid, H5T_NATIVE_INT, &Nx);
	// printf("%s = %d\n", attr, Nx);
	H5Aclose(nxid);
	return Nx;
}

double read_attr_double(hid_t fileid, char *attr)
{
	hid_t nxid;
	double Nx=-1;
	nxid = H5Aopen(fileid, attr, H5P_DEFAULT);
	H5Aread(nxid, H5T_NATIVE_DOUBLE, &Nx);
	// printf("%s = %lf\n", attr, Nx);
	H5Aclose(nxid);
	return Nx;
}


hid_t h5file(char *filename)
{
  	hid_t fileid;
	hid_t nxid, nyid, ntid, lx1id, lx2id, lyid, tmid, epsid;
	htri_t status;
	struct stat buffer;   
	char s;

	double Tmax;
	int Nx, Ny;
	double Lx1, Lx2, Ly;

	s = stat(filename, &buffer)==0;
	// printf("s= %d\n", s);

	if(!s) 
	{
		printf("File does not exist.\n");
		return -1;
	}
	status = H5Fis_hdf5(filename);
	// printf("status = %d\n", status);
	if(status<=0)
	{
		printf("File is not hdf5.\n");
		return -2;
	}
	else
	{
		// printf("File exists and can be opened.\n");
		fileid = H5Fopen(filename, H5F_ACC_RDWR, H5P_DEFAULT);
		if(fileid>0)
		{
			// printf("File is now open.\n");
			return fileid;
		}
	}
	return -3;
}

struct State state_from_hdf5(hid_t fileid)
{
	int ii, jj;
	struct State S;
	double *x, *y;
	int Nx, Ny;
	double Lx1, Lx2, Ly;
	// Open file should be available here
	Nx = read_attr_int(fileid, "Nx");
	Ny = read_attr_int(fileid, "Ny");

	Lx1  = read_attr_double(fileid, "Lx1");
	Lx2  = read_attr_double(fileid, "Lx2");
	Ly   = read_attr_double(fileid, "Ly");

	S.Nx = Nx;
	S.Ny = Ny;
	S.Lx1 = Lx1;
	S.Lx2 = Lx2;
	S.Ly = Ly;
	S.x = (double*) malloc(Nx*sizeof(double));
	S.y = (double*) malloc(Ny*sizeof(double));
	S.xh = (double*) malloc((Nx-1)*sizeof(double));
	S.yh = (double*) malloc((Ny-1)*sizeof(double));
	S.u = (double*) malloc(Nx*(Ny-1)*sizeof(double));
	S.v = (double*) malloc((Nx-1)*Ny*sizeof(double));
	S.unew = (double*) malloc(Nx*(Ny-1)*sizeof(double));

	S.a    = (double*) malloc(Nx*(Ny-1)*sizeof(double));
	S.anew = (double*) malloc(Nx*(Ny-1)*sizeof(double));
	S.b    = (double*) malloc((Nx-1)*(Ny-1)*sizeof(double));
	
	S.uy = (double*) malloc(Nx*sizeof(double));
	S.ay = (double*) malloc(Nx*sizeof(double));
	
	S.Tmax = read_attr_double(fileid, "Tmax");
	S.eps  = read_attr_double(fileid, "eps");
	S.Nt   = read_attr_int(fileid, "Nt");
	S.ncpu   = read_attr_int(fileid, "ncpu");

	S.time = (double*) malloc(6*S.Nt*sizeof(double));
	S.f = S.time + S.Nt;
	
	S.grad  = S.f + S.Nt;
	S.grad1 = S.grad  + S.Nt;
	S.grad2 = S.grad1 + S.Nt;
	S.grad3 = S.grad2 + S.Nt;

	S.diags = (double*) malloc(3*(Ny-1)*S.ncpu*sizeof(double));
	S.W = (double*) malloc(sizeof(double));

    S.debug = (double*) malloc(Nx*Ny*5*sizeof(double));
	
	// for(ii=0;ii<Nx  ;ii++) for(jj=0;jj<Ny-1;jj++) elu(S.u,   jj, ii) = 0;
	// for(ii=0;ii<Nx-1;ii++) for(jj=0;jj<Ny  ;jj++) elv(S.v,   jj, ii) = 0;

	return S;
}

void state_readgrid(hid_t fileid, struct State S)
{
	hid_t xsetid, ysetid, tsetid, fsetid;
	hsize_t dims[1];
	herr_t status;
	int Nt;
	int ii, jj;

	xsetid = H5Dopen(fileid, "/x", H5P_DEFAULT);
	dims[0] = S.Nx;
	status = H5Dread(xsetid, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, S.x);

	ysetid = H5Dopen(fileid, "/y", H5P_DEFAULT);
	dims[0] = S.Ny;
	status = H5Dread(ysetid, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, S.y);

	Nt = read_attr_int(fileid, "Nt");
	tsetid = H5Dopen(fileid, "/t", H5P_DEFAULT);
	dims[0] = Nt;
	status = H5Dread(tsetid, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, S.time);

	fsetid = H5Dopen(fileid, "/f", H5P_DEFAULT);
	dims[0] = Nt;
	status = H5Dread(fsetid, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, S.f);

	for(ii=0;ii<S.Nx-1;ii++) S.xh[ii] = 0.5*(S.x[ii]+S.x[ii+1]);
	for(jj=0;jj<S.Ny-1;jj++) S.yh[jj] = 0.5*(S.y[jj]+S.y[jj+1]);

	return;
}

void state_writecheck(struct Snap snap, int kk, struct State S)
{
		double *us, *vs;
		int ii;
		int Nx = S.Nx, Ny = S.Ny;
		int nn = ++(snap.ilast[0]); // This is the index in us and vs to store
		snap.idcheck[nn] = kk;
		us = snap.ucheck + nn*(Ny-1)*Nx;
		vs = snap.vcheck + nn*(Nx-1)*Ny;
		for(ii=0;ii<(Ny-1)*Nx; ii++) us[ii] = S.u[ii];
		for(ii=0;ii<(Nx-1)*Ny; ii++) vs[ii] = S.v[ii];

		return;
}

void fwdjump(int k1, int k2, int dnc, struct Snap snap, struct State S, int save2disk)
{
	int kk, ii;
	double tmp;
	int Nx = S.Nx;
	int Ny = S.Ny;
	double dt;
	int ncpu = S.ncpu;
	double eps = S.eps;
	double *diags = S.diags;
	double *f = S.f;
	double *time = S.time;
	double *grad1 = S.grad1;

	for( kk=k1+1; kk<k2; kk++)
	{

		dt = time[kk] - time[kk-1];
		fwdstep(S, dt, diags, f[kk], eps, ncpu);
		{double *tmp = S.u; S.u = S.unew; S.unew = tmp;}
		vfromu(S);

		// Now do some post-processing
		for(ii=0;ii<Nx;ii++) S.uy[ii] = (S.x[ii]>=0 && S.x[ii]<=1)?(elu(S.u, 0, ii) + f[kk])/S.yh[0]:0;
		for(ii=0, tmp=0;ii<Nx-1;ii++) tmp += 0.5*(S.uy[ii] + S.uy[ii+1])*(S.x[ii+1]-S.x[ii]);
		grad1[kk] = tmp;
		if (((kk-k1)%dnc == 0) && save2disk) state_writefsnap(snap.fileid, snap.ilast[0]+1, S);
		if ((kk-k1)%dnc == 0) state_writecheck(snap, kk, S);
		if(kk%1000==0) printf("t[%d] = %le     \r", kk, time[kk]);
		fflush(stdout);
	}
	return;
}

int state_readcheck(struct Snap snap, int k2, struct State S)
{
		double *us, *vs;
		int ii;
		int Nx = S.Nx, Ny = S.Ny;
		int nn = (snap.ilast[0]); // This is the last checkpoint index
		int k1 = snap.idcheck[nn] ; // This is the timestamp for last checkpoint

		us = snap.ucheck + nn*(Ny-1)*Nx;
		vs = snap.vcheck + nn*(Nx-1)*Ny;
		for(ii=0;ii<(Ny-1)*Nx; ii++) S.u[ii] = us[ii];
		for(ii=0;ii<(Nx-1)*Ny; ii++) S.v[ii] = vs[ii];
		// (k2+1>k1) means the checkpoint is older than what we want.
		// We want snap at k2, but the last checkpoint is from k1
		// So do a fwdjump registering the intermediate steps as new checkpoints
		if(k2+1>k1)
		{
			fwdjump(k1, k2+1, 1, snap, S, 0); 
		}
		(snap.ilast[0])--; // Pop the last checkpoint as delivered
		return k2;
}

void traverse(struct State S, struct Snap snap, int save2disk)
{
	int ii, jj, kk;
	int Nt = S.Nt;
	int Nx = S.Nx;
	int Ny = S.Ny;

	double *f = S.f;
	double *time = S.time;
	int isnap;
	double W;

	for(ii=0;ii<Nx  ;ii++) for(jj=0;jj<Ny-1;jj++)  elu(S.u, jj, ii) = 0;
	vfromu(S);

	if(save2disk) state_writefsnap(snap.fileid, 0, S);
	state_writecheck(snap, 0, S);


	// omp_set_num_threads(ncpu);
	fwdjump(0, Nt, snap.n2, snap, S, save2disk);

	if(save2disk) state_writefsnap(snap.fileid, snap.ilast[0], S);
	// Initialize a and b
	// For a, ii ranges from 0 to Nx
	for(ii=0;ii<Nx;ii++) for(jj=0;jj<Ny-1;jj++) 
	{
			// double xi = S.yh[jj]/sqrt(S.x[ii]+1e-10);
			// elu(S.a, jj, ii) = -exp(-xi);
			if(jj==0 && S.x[ii]*(1-S.x[ii])>0) elu(S.a, jj, ii) = f[Nt-1];
			else elu(S.a, jj, ii) = 0;
	}

	// time step Nt-1 in S.unew and Nt-2 in S.u
	{double *tmp = S.u; S.u = S.unew; S.unew = tmp;}
	// Calculate v for time step Nt-2
	vfromu(S);
	// Now calculate b at Nt-1 from a[Nt-1], uses a[Nt-1], u[Nt-1] and v[kk-2]
	bfroma(S, f[Nt-1]);
	if (save2disk) state_writebsnap(snap.fileid, snap.ilast[0]+1, S);

	isnap = snap.ilast[0];
	// kk = Nt-2;
	for( kk=Nt-2; kk>0; kk-- )
	// for( kk=Nt-2; kk>Nt-100; kk-- )
	{
		int nc = 0;
		double fkk = f[kk];
		// Step is planned from t[kk+1] to t[kk]
		double dt = (kk<Nt/2) ? (time[kk+1] - time[kk]): (time[Nt-kk-1] - time[Nt-kk-2]);
		double tmp;

		// printf("dt = %le and %le and %le\n", dt, time[1] - time[0], time[Nt-kk] - time[Nt-1-kk]);

		{double *tmp = S.u; S.u = S.unew; S.unew = tmp;}
		// state_readfsnap(snap, kk-1, S);
		// What to replace readfsnap with readcheck
		state_readcheck(snap, kk-1, S);

		// Now S.u contains u[kk-1] and S.unew contains u[kk]
		// Use it to find v[kk-1]
		vfromu(S);
		
		// Ready to take bwdstep to find a[kk] in S.anew from a[kk+1] in S.a
		bwdstep(S, dt, S.diags, fkk, S.eps, S.ncpu);

		// Swap anew and a so that now a[kk] is in S.a
		{double *tmp = S.anew; S.anew = S.a; S.a = tmp; }

		// Find b[kk] from a[kk]
		bfroma(S, fkk); // Check fkk index here

		// Now do some post-processing
		for(ii=0;ii<Nx;ii++) S.ay[ii] = (S.x[ii]>=0 && S.x[ii]<=1)?(elu(S.a, 0, ii) - fkk)/S.yh[0]:0;
		for(ii=0, tmp=0;ii<Nx-1;ii++) tmp += 0.5*(S.ay[ii] + S.ay[ii+1])*(S.x[ii+1]-S.x[ii]);
		S.grad2[kk] = tmp;

		// Computing grad3 is tricky, use S.anew and S.unew
		tmp = 0;
		for(ii=0;ii<Nx;ii++)
		{
			for(jj=0;jj<Ny-1;jj++)
			{
				double up, um, uc;
				double ac;
				double dy;
				uc = elu(S.unew, jj, ii);
				up = (ii<Nx-1)? elu(S.unew, jj, ii+1): elu(S.unew, jj, ii);
				um = (ii>0)   ? elu(S.unew, jj, ii-1): 0;
				ac = elu(S.anew, jj, ii);
				dy = S.y[jj+1] - S.y[jj];
			        tmp += ac*((uc+f[kk]>0)? (uc-um) : (up-uc)) * dy;
			}

		}
		S.grad3[kk] = tmp;

		if(kk%snap.n2 == 0 && save2disk) 
        {
            state_writebsnap(snap.fileid, isnap--, S);
            // printf("Writing isnap = %d\n", isnap);
        }
		if(kk%1000==0) printf("t[%d] = %le     \r", kk, time[kk]);
		fflush(stdout);
	}
	// printf("isnap = %d, kk = %d", isnap, kk);
	if(save2disk) 
    {
        state_writebsnap(snap.fileid, isnap--, S);
        // printf("Writing isnap = %d\n", isnap);
    }

	for(kk=0;kk<Nt;kk++) S.grad[kk] = S.grad1[kk] - S.grad2[kk] - S.eps*S.grad3[kk];
	S.grad[0] = 0;
	S.grad[Nt-1] = 0;
	{
		double integral = trapz(S.time, S.grad, Nt);
		for(kk=1;kk<Nt-1;kk++) S.grad[kk] -= integral;
	}
	S.W[0] = dot2(S.grad1, S.f, S.time, Nt);
}

void state_writef(hid_t fileid, struct State S)
{
	hid_t g1setid, g2setid, g3setid, gsetid, fsetid, Wid;
	herr_t status;

	fsetid = H5Dopen(fileid, "/f", H5P_DEFAULT);
	status = H5Dwrite(fsetid, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, S.f);
	H5Dclose(fsetid);

	g1setid = H5Dopen(fileid, "/Grad1", H5P_DEFAULT);
	status = H5Dwrite(g1setid, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, S.grad1);
	H5Dclose(g1setid);

	g2setid = H5Dopen(fileid, "/Grad2", H5P_DEFAULT);
	status = H5Dwrite(g2setid, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, S.grad2);
	H5Dclose(g2setid);

	g3setid = H5Dopen(fileid, "/Grad3", H5P_DEFAULT);
	status = H5Dwrite(g3setid, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, S.grad3);
	H5Dclose(g3setid);

	gsetid = H5Dopen(fileid, "/Grad", H5P_DEFAULT);
	status = H5Dwrite(gsetid, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, S.grad);
	H5Dclose(gsetid);

	Wid = H5Aopen(fileid, "W", H5P_DEFAULT);
	H5Awrite(Wid, H5T_NATIVE_DOUBLE, S.W);
	H5Aclose(Wid);
}
