
#ifndef FLATPLATE_H
#define FLATPLATE_H
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <hdf5.h>
#include <omp.h>
#include <sys/stat.h>
#include <string.h>

#define elu(u,j,i) (u[(Ny-1)*(i)+(j)])
#define elv(v,j,i) (v[Ny*(i)+(j)])
#define pi M_PI

// Definition of dgtsv_ from LAPACK
void dgtsv_(int *N, int *nrhs, double *dl, double *d, double *du, double *b, int *ldb, int *info);

struct Snap 
{
	hid_t fileid;

	double *ucheck;
	double *vcheck;
	int *idcheck;
	int n1, n2;    // Checkpoints in each level
	int *ilast;     // The index of the last checkpoint
	// We must have n1 + n2 < ncheck and n1*n2 > Nt
};

struct State
{
	int Nx;
	int Ny;
	double Lx1;
	double Lx2;
	double Ly;
	double Tmax;
	double *x;
	double *y;
	double *xh;
	double *yh;

	double *u;
	double *v;
	double *a;
	double *b;
	double *unew;
	double *anew;
	double *uy;
	double *ay;

	long int Nt;
	double *time;
	double *f;
	double *diags;

	double *grad1;
	double *grad2;
	double *grad3;
	double *grad;

	double eps;
	double *W;
	int ncpu;

    double *debug;
};

double dot2(double *a, double *b, double *t, int Nt);
double grid(double kk, int Nt, double L1, double L2, char type);
struct State state_new(int Nx, int Ny, double Lx1, double Lx2, double Ly);
double trapz(double *time, double *f, int Nt);
void state_writefsnap(hid_t fileid, int nn, struct State S);
void state_readfsnap(hid_t fileid, int nn, struct State S);
void state_writebsnap(hid_t fileid, int nn, struct State S);
void state_readbsnap(hid_t fileid, int nn, struct State S);
hid_t inith5(char *fname, struct State S, int Nt, double *t, double *f);
void vfromu(struct State S);
void bfroma(struct State S, double fkk);
void fwdstep(struct State S, double dt, double *diags, double fkk, double eps, int ncpu);
void bwdstep(struct State S, double dt, double *diags, double fkk, double eps, int ncpu);
int read_attr_int(hid_t fileid, char *attr);
double read_attr_double(hid_t fileid, char *attr);
hid_t h5file(char *filename);
struct State state_from_hdf5(hid_t fileid);
void state_readgrid(hid_t fileid, struct State S);
void state_writecheck(struct Snap snap, int kk, struct State S);
void fwdjump(int k1, int k2, int dnc, struct Snap snap, struct State S, int save2disk);
int state_readcheck(struct Snap snap, int k2, struct State S);
void traverse(struct State S, struct Snap snap, int save2disk);
void state_writef(hid_t fileid, struct State S);
#endif
