#include "flatplate.h"

double dot(double *a, double *b, double *t, int Nt)
{
	double sum = 0;
	int kk;
	// for(kk=1;kk<Nt-1;kk++) sum += a[kk]*b[kk]*(t[kk+1]-t[kk]);
	for(kk=1;kk<Nt-1;kk++) sum += a[kk]*b[kk];
	return sum;
}

double L1(double *a, int Nt)
{
	int kk;
	double l1 = 0;
	for(kk=1;kk<Nt-1;kk++) l1 = (fabs(a[kk])>l1) ? fabs(a[kk]) : l1;
	return l1;
}

double integ(double *a, double *t, int Nt)
{
	double sum = 0;
	int kk;
	for(kk=1;kk<Nt-1;kk++) sum += a[kk]*(t[kk+1]-t[kk]);
	return sum;
}
int main(int argc, char **argv)
{
	int Nx, Ny, Nt;
	int ii, jj, kk;
	struct State S;
	struct Snap snap;
	int ncheck;

	double *f0, *g0, *v1, *w1;
	double a1 = 1e-3;
	double b1 = 1e-5;

	double A11, A12, A21, A22;
	double r1, r2;
	double c1, c2;
	double L1Grad, L2Grad, W;
	int iter;
	int nosave2disk=0;
	int save2disk=1;
	int niter = 10;
	char h5fname[1000] = "snaps.hdf5";

	if(argc>1)
	{
		niter = atoi(argv[1]);
	}
	printf("Executing %d iterations of steepest descent.\n", niter);
	// printf("argc = %d\n", argc);
	if(argc>2) 
	{
		printf("File: %s\n", argv[2]);
		strcpy(h5fname, argv[2]);
	}
	// else snap.fileid = h5file("snaps.hdf5");
	printf("Reading from hdf5 file: %s\n", h5fname);
	snap.fileid = h5file(h5fname);
	if(snap.fileid<=0)
	{
		printf("Invalid hdf5 file or file does not exist. Quiting.\n");
		return 1;
	}
	S = state_from_hdf5(snap.fileid);
	snap.n1 = read_attr_int(snap.fileid, "n1");
	snap.n2 = read_attr_int(snap.fileid, "n2");
	
	Ny = S.Ny; Nx = S.Nx; Nt = S.Nt;

	ncheck = snap.n1 + snap.n2;
	snap.idcheck = (int*)   malloc(ncheck*sizeof(int));
	snap.ilast = (int*) malloc(sizeof(int));
	snap.ucheck = (double*) malloc(ncheck*(Nx*(Ny-1) + Ny*(Nx-1))*sizeof(double));

	if (snap.ucheck == NULL) 
	{
		printf("Memory for checkpoints: %f MB, too much. Saving to disk ...\n", 
				((double)ncheck)*(Nx*(Ny-1) + Ny*(Nx-1))/(1024*1024));
		return -2;
	}
	else 
	{
		printf("Using %f MB of memory for checkpoints\n", ((double)ncheck)*8*(Nx*(Ny-1) + Ny*(Nx-1))/(1024*1024)); 
		snap.vcheck = snap.ucheck + ncheck*Nx*(Ny-1);
		snap.ilast[0] = -1; // There are no checkpoints to begin with
	}

	f0 = (double*)malloc(Nt*sizeof(double));
	g0 = (double*)malloc(Nt*sizeof(double));
	v1 = (double*)malloc(Nt*sizeof(double));
	w1 = (double*)malloc(Nt*sizeof(double));

	// Now we need to read S.x, S.y, time and f from fileid
	state_readgrid(snap.fileid, S);
	H5Fflush(snap.fileid, H5F_SCOPE_GLOBAL);
	H5Fclose(snap.fileid);

	for(iter=0;iter<niter;iter++)
	{
		for(kk=0;kk<Nt;kk++) f0[kk] = S.f[kk];
		traverse(S, snap, nosave2disk);
		for(kk=0;kk<Nt;kk++) g0[kk] = S.grad[kk];
		L1Grad = L1(g0, Nt);
		L2Grad = sqrt(dot2(g0, g0, S.time, Nt));
		// W = dot2(S.grad1, S.f, S.time, Nt);

		printf("W = %le\t, Linf(grad) = %le, L2(grad) = %le\n", S.W[0], L1Grad, L2Grad);

		for(kk=1;kk<Nt-1;kk++) S.f[kk] = f0[kk] - a1*g0[kk];
		traverse(S, snap, nosave2disk);
		for(kk=0;kk<Nt;kk++) v1[kk] = (S.grad[kk]-g0[kk])/a1;

		for(kk=1;kk<Nt-1;kk++) S.f[kk] = f0[kk] - b1*v1[kk];
		traverse(S, snap, nosave2disk);
		for(kk=0;kk<Nt;kk++) w1[kk] = (S.grad[kk]-g0[kk])/b1;

		r1 = dot(g0, g0, S.time, Nt);
		r2 = dot(g0, v1, S.time, Nt);

		A11 = r2;
		A21 = dot(v1, v1, S.time, Nt);
		A12 = dot(w1, g0, S.time, Nt);
		A22 = dot(w1, v1, S.time, Nt);

		/*
		r1 = dot2(g0, g0, S.time, Nt);
		r2 = dot2(g0, v1, S.time, Nt);

		A11 = r2;
		A21 = dot2(v1, v1, S.time, Nt);
		A12 = dot2(w1, g0, S.time, Nt);
		A22 = dot2(w1, v1, S.time, Nt);
		*/

		// printf("A11 = %le\t A12 = %le\t \t r1 = %le\n", A11, A12, r1);
		// printf("A21 = %le\t A22 = %le\t \t r2 = %le\n", A21, A22, r2);

		// a x + b y = r1   x = d r1 - b r2
		// c x + d y = r2   y =-c r1 + a r2
		c1 =( A22*r1 - A12*r2)/(A11*A22-A12*A21);
		c2 =(-A21*r1 + A11*r2)/(A11*A22-A12*A21);

		printf("c1 = %le \t c2 = %le\n", c1, c2);

		for(kk=1;kk<Nt;kk++) S.f[kk] = f0[kk] + c1*g0[kk] + c2*v1[kk];
	
		snap.fileid = h5file(h5fname);
		state_writef(snap.fileid, S);
		H5Fclose(snap.fileid);
	
	}
	
	// Write the components of the gradients
	snap.fileid = h5file(h5fname);
	fwdjump(0, Nt, snap.n2, snap, S, save2disk);
	H5Fclose(snap.fileid);
	return 0;
}

