import h5py as h5
import numpy as np
import matplotlib.pyplot as plt
import sys

def grid(N, L1, L2, typ=1):
    s = np.linspace(0,1,N)
    L = L2-L1
    if(typ==1):
        return L1 + L*s
    if(typ==2):
        return L1+L*s**2*(3-2*s)
    if(typ==3):
        return L1+L*s**3*(10-15*s+6*s**2)
    if(typ==4):
        return L1+L*s**4*(35-84*s+70*s**2-20*s**3)
    if(typ==-1):
        lam = 5
        fac = 1.0/(1+np.exp(-lam)/lam - 1.0/lam);
        return L1 + L*fac*(s + np.exp(-lam*s)/lam - 1.0/lam)
    else:
        return L1+L*s

if len(sys.argv) >= 2:
    ceps = sys.argv[1]
    eps = float(ceps)
else:
    eps = 1.0
print('eps = ', eps)

# Nx = 401
Nx = 101
Ny = 401
Nt = 32001

Lx1 = -0.5
Lx2 = 1.5
Ly = 8.0
Tmax = 1.0
ixtyp = 3
ittyp = 3

y = grid(Ny,0, Ly, typ=-1)
# y = grid(Ny,0, Ly, typ=1)
i1 = int((Nx-1)/4 + 1)
i2 = int((Nx-1)*3/4 + 1)
x = np.zeros((Nx,))
x[:i1] = grid(i1, Lx1, 0, typ=ixtyp)
x[i1-1:i2] = grid(i2-i1+1, 0, 1, typ=ixtyp)
x[i2-1:] = grid(Nx-i2+1, 1, Lx2, typ=ixtyp)

dx = x[1:]-x[:-1]
dxmin = dx.min()

Nt = int(eps*int(1.0/dxmin/100)*150)+1
Nt = max(Nt, 401)
t = grid(Nt,0, 1, typ=ittyp)
# f = 6*t*(1-t)
f = 0*t+1
# f = (t*(1-t))**0.25
# f /= np.trapz(f,x=t)

f[0] = 0
f[-1] = 0

n1 = int(np.ceil((Nt*1.0)**0.5));
n2 = int(np.ceil(Nt*1.0/n1));
ncheck = n1 + n2;
ncpu = 64;

# Space for u, v, a and b
print("Space for {0} snapshots: {1} MB".format(ncheck, ncheck*((Nx*(Ny-1) + Ny*(Nx-1))*16.0/(1024*1024))))

h5fname = 'snaps.hdf5'
f5 = h5.File(h5fname, mode='a', libver='earliest')

f5.attrs.create('eps',  eps,  dtype='d')
f5.attrs.create('Nx',   Nx,   dtype='i')
f5.attrs.create('Ny',   Ny,   dtype='i')
f5.attrs.create('Nt',   Nt,   dtype='i')
f5.attrs.create('Lx1',  Lx1,  dtype='d')
f5.attrs.create('Lx2',  Lx2,  dtype='d')
f5.attrs.create('Ly',   Ly,   dtype='d')
f5.attrs.create('Tmax', Tmax, dtype='d')
f5.attrs.create('L1min', Tmax, dtype='d')
f5.attrs.create('n1', n1, dtype='i')
f5.attrs.create('n2', n2, dtype='i')
f5.attrs.create('ncpu', ncpu, dtype='i')
f5.attrs.create('W', np.zeros((1,)), dtype='d')

f5.create_dataset('y', data=y)
f5.create_dataset('x', data=x)
f5.create_dataset('t', data=t)
f5.create_dataset('f', data=f)
f5.create_dataset('fmin', data=f)

u = np.zeros((ncheck,Nx,Ny-1))
v = np.zeros((ncheck,Nx-1,Ny))
b = np.zeros((ncheck,Nx-1,Ny-1))
uy = np.zeros((ncheck,Nx))

f5.create_dataset('u', data=u, chunks=(1,Nx,Ny-1))
f5.create_dataset('v', data=v, chunks=(1,Nx-1,Ny))
f5.create_dataset('a', data=u, chunks=(1,Nx,Ny-1))
f5.create_dataset('b', data=b, chunks=(1,Nx-1,Ny-1))
f5.create_dataset('uys', data=uy)
f5.create_dataset('ays', data=uy)

f5.create_dataset('Grad1', shape=(Nt,), dtype='d');
f5.create_dataset('Grad2', shape=(Nt,), dtype='d');
f5.create_dataset('Grad3', shape=(Nt,), dtype='d');
f5.create_dataset('Grad', shape=(Nt,), dtype='d');

f5.close()
