import numpy as np
import matplotlib.pyplot as plt
import scipy.sparse as ss
import scipy.sparse.linalg as ssl
import h5py as h5
import scipy.interpolate as si

###############################################################################
### Construct Dyyyy
### Size: n x n
###############################################################################
def D4y(Nx, Ny, dy, inid):
    rows = []
    cols = []
    data = []

    c = 1/dy**4

    jj = 0
    for ii in range(Nx):
        row = ii + Nx*0
        col0 = ii + Nx*0
        col1 = ii + Nx*1
        col2 = ii + Nx*2
        if ii in inid:
            data0 =  7*c
        else:
            data0 = 5*c
        data1 = -4*c
        data2 =  1*c
        rows += [row, row, row]
        cols += [col0, col1, col2]
        data += [data0, data1, data2]

    jj = 1
    for ii in range(Nx):
        row = ii + Nx*1
        col1 = ii + Nx*0
        col2 = ii + Nx*1
        col3 = ii + Nx*2
        col4 = ii + Nx*3
        data1 = -4*c
        data2 =  6*c
        data3 = -4*c
        data4 =  1*c
        rows += [row, row, row, row]
        cols += [col1, col2, col3, col4]
        data += [data1, data2, data3, data4]

        
    for jj in range(2,Ny-3):
        for ii in range(Nx):
            row = ii + Nx*jj
            col1 = ii + (jj-2)*Nx
            col2 = ii + (jj-1)*Nx
            col3 = ii + (jj+0)*Nx
            col4 = ii + (jj+1)*Nx
            col5 = ii + (jj+2)*Nx
            data1 =  1*c
            data2 = -4*c
            data3 =  6*c
            data4 = -4*c
            data5 =  1*c
            rows += [row,  row, row, row, row]
            cols += [col1, col2, col3, col4, col5]
            data += [data1, data2, data3, data4, data5]

    jj = Ny-3
    for ii in range(Nx):
        row = ii + Nx*jj
        col1 = ii + Nx*(jj+1)
        col2 = ii + Nx*(jj+0)
        col3 = ii + Nx*(jj-1)
        col4 = ii + Nx*(jj-2)
        data1 = -2*c
        data2 =  5*c
        data3 = -4*c
        data4 =  1*c
        rows += [row, row, row, row]
        cols += [col1, col2, col3, col4]
        data += [data1, data2, data3, data4]

    jj = Ny-2
    for ii in range(Nx):
        row = ii + Nx*jj
        col2 = ii + Nx*(jj+0)
        col3 = ii + Nx*(jj-1)
        col4 = ii + Nx*(jj-2)
        data2 =  2*c
        data3 = -4*c
        data4 =  2*c
        rows += [row, row, row]
        cols += [col2, col3, col4]
        data += [data2, data3, data4]

    Dyyyy = ss.coo_matrix((data, (rows, cols)))
    return Dyyyy

###############################################################################
### Construct Dxxxx
### Size: n x n
###############################################################################
def D4x(Nx, Ny, dx):
    rows = []
    cols = []
    data = []

    c = 1/dx**4
        
    for jj in range(Ny-1):
        ii = 0
        row = ii + Nx*jj
        col3 = ii+0 + (jj)*Nx
        col4 = ii+1 + (jj)*Nx
        col5 = ii+2 + (jj)*Nx
        data3 =  6*c
        data4 = -8*c
        data5 =  2*c
        rows += [row,  row, row]
        cols += [col3, col4, col5]
        data += [data3, data4, data5]

        ii = 1
        row = ii + Nx*jj
        col2 = ii-1 + (jj)*Nx
        col3 = ii+0 + (jj)*Nx
        col4 = ii+1 + (jj)*Nx
        col5 = ii+2 + (jj)*Nx
        data2 = -4*c
        data3 =  7*c
        data4 = -4*c
        data5 =  1*c
        rows += [row,   row,   row,   row  ]
        cols += [col2,  col3,  col4,  col5 ]
        data += [data2, data3, data4, data5]

        for ii in range(2,Nx-2):
            row = ii + Nx*jj
            col1 = ii-2 + (jj)*Nx
            col2 = ii-1 + (jj)*Nx
            col3 = ii+0 + (jj)*Nx
            col4 = ii+1 + (jj)*Nx
            col5 = ii+2 + (jj)*Nx
            data1 =  1*c
            data2 = -4*c
            data3 =  6*c
            data4 = -4*c
            data5 =  1*c
            rows += [row,  row, row, row, row]
            cols += [col1, col2, col3, col4, col5]
            data += [data1, data2, data3, data4, data5]

        ii = Nx-2
        row = ii + Nx*jj
        col1 = ii-2 + (jj)*Nx
        col2 = ii-1 + (jj)*Nx
        col3 = ii+0 + (jj)*Nx
        col4 = ii+1 + (jj)*Nx
        data1 =  1*c
        data2 = -4*c
        data3 =  7*c
        data4 = -4*c
        rows += [row,  row, row, row]
        cols += [col1, col2, col3, col4]
        data += [data1, data2, data3, data4]

        ii = Nx-1
        row = ii + Nx*jj
        col1 = ii-2 + (jj)*Nx
        col2 = ii-1 + (jj)*Nx
        col3 = ii+0 + (jj)*Nx
        data1 =  2*c
        data2 = -8*c
        data3 =  6*c
        rows += [row,  row, row]
        cols += [col1, col2, col3,  ]
        data += [data1, data2, data3]
    Dxxxx = ss.coo_matrix((data, (rows, cols)))
    return Dxxxx

###############################################################################
### Construct D2y
### Size: (Ny-1)Nx x (Ny-1)Nx
###############################################################################
def D2yfun(Nx, Ny, dy, inid):
    rows = []
    cols = []
    data = []

    c = 1/dy**2

    jj = 0 # This is y=0 
    for ii in range(Nx):
        row = ii + Nx*jj
        col0 = ii + Nx*jj
        if ii in inid:
            data0 = +2*c
        else:
            data0 = 0
        rows += [row ]
        cols += [col0]
        data += [data0]

    jj = 1 # This is y=dy, jj-1 is the diagonal col 
    for ii in range(Nx):
        row = ii + Nx*jj
        col0 = ii + Nx*(jj-1)   # diag
        col1 = ii + Nx*(jj-0)   # super
        data0 = -2*c
        data1 = +1*c

        rows += [row,  row]
        cols += [col0, col1]
        data += [data0, data1]
    
        
    for jj in range(2,Ny-1):
        for ii in range(Nx):
            row = ii + Nx*(jj)
            col0 = ii + (jj-2)*Nx   # sub
            col1 = ii + (jj-1)*Nx   # diag
            col2 = ii + (jj-0)*Nx   # super

            data0 =  1*c
            data1 = -2*c
            data2 =  1*c

            rows += [row,   row,   row]
            cols += [col0,  col1,  col2]
            data += [data0, data1, data2]

    D2y = ss.coo_matrix((data, (rows, cols)))
    return D2y    

###############################################################################
### Construct D2x
### Size: (Ny-1)Nx x (Ny-1)(Nx-3)
###############################################################################
def D2xfun(Nx, Ny, dx):
    rows = []
    cols = []
    data = []

    c = 1/dx**2
        
    for jj in range(Ny-1):
        for ii in range(1,Nx-1):
            row = (ii-1) + (Nx-2)*jj
            col0 = ii-1 + (jj)*Nx
            col1 = ii+0 + (jj)*Nx
            col2 = ii+1 + (jj)*Nx
            data0 = +1*c
            data1 = -2*c
            data2 = +1*c

            rows += [row, row,  row]
            cols += [col0, col1, col2]
            data += [data0, data1, data2]

    D2x = ss.coo_matrix((data, (rows, cols)))
    return D2x

###############################################################################
### Construct Dx
### Size: (Ny-1)Nx x (Ny-1)(Nx-1)
###############################################################################
def Dxfun(Nx, Ny, dx):
    rows = []
    cols = []
    data = []

    c = 1/dx
        
    for jj in range(Ny-1):
        for ii in range(Nx-1):
            row = ii + (Nx-1)*jj
            col1 = ii+0 + (jj)*Nx
            col2 = ii+1 + (jj)*Nx
            data1 = -1*c
            data2 =  1*c
            rows += [row,  row]
            cols += [col1, col2]
            data += [data1, data2]

    Dx = ss.coo_matrix((data, (rows, cols)))
    return Dx

###############################################################################
### Construct Dy
### Size: (Ny-1)Nx x (Ny-1)Nx
###############################################################################
def Dyfun(Nx, Ny, dy, inid):
    rows = []
    cols = []
    data = []

    c = 1/dy

    jj = 0
    for ii in range(Nx):
        row = ii + Nx*0
        col0 = ii + Nx*0
        if ii in inid:
            data0 = +c
        else:
            data0 = 0
        rows += [row ]
        cols += [col0]
        data += [data0]

        
    for jj in range(1,Ny-1):
        for ii in range(Nx):
            row = ii + Nx*(jj)
            col1 = ii + (jj-0)*Nx
            col2 = ii + (jj-1)*Nx

            data1 =  1*c
            data2 = -1*c

            rows += [row,  row]
            cols += [col1, col2]
            data += [data1, data2]

    Dy = ss.coo_matrix((data, (rows, cols)))
    return Dy



###############################################################################
### Construct DyAxy from ax
### Size: nxn from (Ny-1, Nx)
###############################################################################
def AxMat(ax, Nx, Ny, dy):
    rows = []
    cols = []
    data = []

    c = 1/dy**2
    jj = 0
    for ii in range(Nx):
        row = ii + Nx*jj
        col1 = ii + (jj+0)*Nx 
        col2 = ii + (jj+1)*Nx
        data1 = -(ax[jj,ii] + ax[jj+1,ii])*c
        data2 = ax[jj+1,ii]*c
        rows += [row, row]
        cols += [col1, col2]
        data += [data1, data2]

    for jj in range(1,Ny-2):
        for ii in range(Nx):
            # Here y_j = (jj+1)*dy
            row = ii + Nx*jj
            col1 = ii + (jj-1)*Nx   # (j-1) dy  sub
            col2 = ii + (jj+0)*Nx   # j dy      diagonal
            col3 = ii + (jj+1)*Nx   # (j+1) dy  super
            data1 =  ax[jj,ii]*c
            data2 = -(ax[jj,ii] + ax[jj+1,ii])*c
            data3 =  ax[jj+1,ii]*c
            rows += [row,  row, row]
            cols += [col1, col2, col3]
            data += [data1, data2, data3]

    jj = Ny-2
    for ii in range(Nx):
        row = ii + Nx*jj
        col1 = ii + (jj-1)*Nx 
        col2 = ii + (jj+0)*Nx
        data1 = ax[jj,ii]*c
        data2 = -(ax[jj,ii] + ax[jj,ii])*c
        rows += [row, row]
        cols += [col1, col2]
        data += [data1, data2]

    DyAxy = ss.coo_matrix((data, (rows, cols)))
    return DyAxy

###############################################################################
### Construct DxAyy + DyAxx from ax
### Size: n x n from (Ny-1, Nx)
### But, this is constructed in steps
###############################################################################

# First the Dx, has size (Ny-1)(Nx-1) x (Ny-1)Nx
def Dx4Ay(Nx, Ny, dx):
    rows = []
    cols = []
    data = []

    jj = 0
    for ii in range(Nx-1):
        row = ii + 0*(Nx-1)
        col1 = ii+1 + jj*Nx
        col2 = ii   + jj*Nx
        data1 =  0.5/dx
        data2 = -0.5/dx
        rows += [row,  row]
        cols += [col1, col2]
        data += [data1, data2]

    for jj in range(Ny-2):
        for ii in range(Nx-1):
            row = ii + (jj+1)*(Nx-1)
            col1 = ii + 0 + (jj+0)*Nx
            col2 = ii + 1 + (jj+0)*Nx
            col3 = ii + 0 + (jj+1)*Nx
            col4 = ii + 1 + (jj+1)*Nx
            data1 = -0.5/dx
            data2 = +0.5/dx
            data3 = -0.5/dx
            data4 = +0.5/dx
            rows += [row, row, row, row]
            cols += [col1, col2, col3, col4]
            data += [data1, data2, data3, data4]

    Dx = ss.coo_matrix((data, (rows, cols)))
    return Dx

# Next the Dy, has size (Ny-1)(Nx-1) x (Ny-1)Nx
def Dy4Ay(Nx, Ny, dy):
    rows = []
    cols = []
    data = []

    jj = 0
    for ii in range(Nx-1):
        row = ii + 0*(Nx-1)
        col1 = ii+1 + jj*Nx
        col2 = ii   + jj*Nx
        data1 =  0.5/dy
        data2 =  0.5/dy
        rows += [row,  row]
        cols += [col1, col2]
        data += [data1, data2]

    for jj in range(Ny-2):
        for ii in range(Nx-1):
            row = ii + (jj+1)*(Nx-1)
            col1 = ii + 0 + (jj+0)*Nx
            col2 = ii + 1 + (jj+0)*Nx
            col3 = ii + 0 + (jj+1)*Nx
            col4 = ii + 1 + (jj+1)*Nx
            data1 = -0.5/dy
            data2 = -0.5/dy
            data3 = +0.5/dy
            data4 = +0.5/dy
            rows += [row, row, row, row]
            cols += [col1, col2, col3, col4]
            data += [data1, data2, data3, data4]

    Dy = ss.coo_matrix((data, (rows, cols)))
    return Dy

def vec2rhs(vec, Nx, Ny):
    psi = vec.reshape((Ny-1, Nx))
    lap = 0*psi
    lap[1:-1,1:-1] = (psi[2:,1:-1] - 2*psi[1:-1,1:-1] + psi[:-2,1:-1])/dy**2 \
                   + (psi[1:-1,2:] - 2*psi[1:-1,1:-1] + psi[1:-1,:-2])/dx**2

    lap[0,1:-1] =  (psi[1,1:-1] - 2*psi[0,1:-1])/dy**2 \
                   + (psi[0,2:] - 2*psi[0,1:-1] + psi[0,:-2])/dx**2

    lap[-1,1:-1] = (psi[-1,2:] - 2*psi[-1,1:-1] + psi[-1,:-2])/dx**2

    lap[1:-1,0] = (psi[2:,0] - 2*psi[1:-1,0] + psi[:-2,0])/dy**2
    lap[1:-1,-1] = (psi[2:,-1] - 2*psi[1:-1,-1] + psi[:-2,-1])/dy**2

    lap[0,0] =  (psi[1,0] - 2*psi[0,0])/dy**2 
    lap[0,-1] =  (psi[1,-1] - 2*psi[0,-1])/dy**2

    rhs = (psi-lap).ravel()
    return rhs

def vec2rhsall(vecin, Nx, Ny):
    vec = vecin.copy()
    n, nev = vec.shape
    rhs = np.zeros(vec.shape)
    for ii in range(nev):
        rhs[:,ii] = vec2rhs(vec[:,ii], Nx, Ny)
        dot = (vec[:,ii].T@rhs[:,ii])
        vec[:,ii] /= dot**0.5
        rhs[:,ii] /= dot**0.5
        for jj in range(ii+1,nev):
            rhs[:,jj] = vec2rhs(vec[:,jj], Nx, Ny)
            dot = (vec[:,ii].T@rhs[:,jj])
            vec[:,jj] -= dot*vec[:,ii]
            # rhs[:,jj] -= dot*rhs[:,ii]
    return vec,rhs

###############################################################################
### Program starts here
###############################################################################


fname = '../c/data/snap_eps01.0.hdf5'
f5 = h5.File(fname, 'r')
Lx15 = f5.attrs['Lx1']
Lx25 = f5.attrs['Lx2']
Ly5  = f5.attrs['Ly']

Nx5 = f5.attrs['Nx']
Ny5 = f5.attrs['Ny']
n1 = f5.attrs['n1']
n2 = f5.attrs['n2']
eps = f5.attrs['eps']

x5 = f5['x'][:]
y5 = f5['y'][:]
u5 = f5['u'][:]
a5 = f5['a'][:]
t = f5['t'][:]
f = f5['f'][:]

f5.close()

yh5 = 0.5*(y5[1:]+y5[:-1])
xh5 = 0.5*(x5[1:]+x5[:-1])
tkk = np.hstack( ( t[range(0,n1*n2,n2)], t[-1]))
fkk = np.hstack( ( f[range(0,n1*n2,n2)], f[-1]))

id0 = np.where(np.abs(x5)<1e-7)[0][0]
id1 = np.where(np.abs(x5-1)<1e-7)[0][0]

a5 = eps*(a5 + u5)
a5[:, 0,:] = a5[:, 1,:]
a5[:,-1,:] = a5[:,-2,:]

nim = 7
nip = 8
for ij in range(-nim+1, nip):
    a5[:, id0-ij, :] = a5[:, id0-nim,:] + (x5[id0-ij]-x5[id0-nim])*(a5[:,id0+nip,:] - a5[:,id0-nim,:])/(x5[id0+nip]-x5[id0-nim])
    a5[:, id1-ij, :] = a5[:, id1-nim,:] + (x5[id1-ij]-x5[id1-nim])*(a5[:,id1+nip,:] - a5[:,id1-nim,:])/(x5[id1+nip]-x5[id1-nim])

# Generic grid library
Nx = 201
Ny = 401
Lx1 = x5[1]
Lx2 = x5[-2]
Ly1 = 0
Ly2 = y5[-2]

dx = (Lx2-Lx1)/(Nx-1)
dy = (Ly2-Ly1)/(Ny-1)
x = np.arange(Nx)*dx + Lx1
y = np.arange(Ny)*dy + Ly1

xh = 0.5*(x[1:] + x[:-1])
yh = 0.5*(y[1:] + y[:-1])

Xu,Yu = np.meshgrid(x,  yh)
Xv,Yv = np.meshgrid(xh, y )
X, Y  = np.meshgrid(x,  y )
Xp,Yp = np.meshgrid(xh, yh)

tol = 0
nev = 12
n = (Ny-1)*Nx
inid = np.where(x*(1-x)>=0)[0]
onid = np.where(x*(1-x)<0)[0]


allevals = np.zeros((n1, nev))

f5 = h5.File(fname, 'r+')
if 'uvecs' in f5.keys():
    del f5['uvecs']
if 'evals' in f5.keys():
    del f5['evals']

f5.create_dataset('uvecs', (n1, nev, Ny-1, Nx), dtype=np.float64)
f5.create_dataset('evals', data=allevals)
f5.close()

psi = np.random.random((Ny-1,Nx))
# psi = (1-np.cos(Y*np.pi/2/Ly2))*np.sin((X-Lx1)*np.pi/(Lx2-Lx1))
# psi = psi[1:,:]
vec = psi.ravel()
# vec = np.random.random((n, nev))*dx*dy

# raise SystemExit 

nplot = 100
for nn in range(n1):
# for nn in range(12,14):

    # ax = -10*np.exp(-5*Yu)*(1-np.cos((Xu-Lx1)*2*np.pi/(Lx2-Lx1))) # (Ny-2) x Nx
    # ay = 10*np.exp(-5*Yp)*np.sin((Xp-Lx1)*2*np.pi/(Lx2-Lx1))       # (Ny-1) x (Nx-1)

    print(" *** nn = {0}  of {1} *** ".format(nn, n1), flush=True)
    ax5 = (a5[nn,1:,:] - a5[nn,:-1,:])/np.tile(x5[1:] - x5[:-1], (Ny5-1,1)).T
    ay5 = (a5[nn,:,1:]-a5[nn,:,:-1])/(yh5[1:]-yh5[:-1])

    axfun = si.RectBivariateSpline(xh5, yh5, ax5)
    ax = axfun(x, yh).T

    ayfun = si.RectBivariateSpline(x5, y5[1:-1], ay5)
    ay = ayfun(xh, yh).T

    small = 1e-2 # upto 1e-4
    # Dyyyy = D4y(Nx, Ny, dy, inid)
    # Dxxxx = D4x(Nx, Ny, dy)
    D2y = D2yfun(Nx, Ny, dy, inid)
    D2x = D2xfun(Nx, Ny, dx)
    Dyyyy = D2y.T@D2y
    Dxxxx = D2x.T@D2x
    DyAxy = AxMat(ax, Nx, Ny, dy)
    Dx = Dx4Ay(Nx, Ny, dx)
    Dy = Dy4Ay(Nx, Ny, dy)
    Ay = ss.diags([ay.ravel()], [0])
    DAy = 0.5*(Dx.T@Ay@Dy + Dy.T@Ay@Dx)
    Mat = Dyyyy + small*Dxxxx - DyAxy - DAy

    Dxmat = Dxfun(Nx, Ny, dx)
    Dymat = Dyfun(Nx, Ny, dy, inid)
    RMat = ss.eye(n) + Dxmat.T@Dxmat + Dymat.T@Dymat
    # Mat = 0.5*(Mat + Mat.T)
    # Matlu = ssl.splu(Mat.tocsc())
    
    # evals, evecs = ssl.eigsh(Mat, M=RMat, which='SA', v0=vec)
    evals, evecs = ssl.eigsh(Mat, M=RMat, which='LA', sigma=-12e-0, v0=vec, k=nev, tol=tol)
    print('evals = ', evals.real)

    allevals[nn,:] = evals

    f5 = h5.File(fname, 'r+')
    for ii in range(nev):
        f5['uvecs'][nn,ii,:,:] = evecs[:,ii].reshape((Ny-1,Nx))
    f5['evals'][nn,:] = evals
    f5.close()
    
    vec = evecs[:,0]

    psi = np.zeros((Ny+1,Nx))
    psi[2:,:] = vec.reshape((Ny-1, Nx))
    psi[2:,:] = evecs[:,0].reshape((Ny-1, Nx)).real
    psi[1,:] = 0
    psi[0,inid] = psi[2,inid]
    psi[0,onid] = -psi[2,onid]

    psiy = (psi[1:,:] - psi[:-1,:])/dy
    psix = (psi[:,1:] - psi[:,:-1])/dx
    psiyy = (psi[2:,:] - 2*psi[1:-1,:] + psi[:-2,:])/dy**2
    psixx = (psi[:,2:] - 2*psi[:,1:-1] + psi[:,:-2])/dx**2

    plt.clf()
    # plt.pcolor(psix)
    plt.pcolor(x5, y5, a5[nn,:,:].T)
    # plt.pcolor(ax)
    plt.colorbar()
    plt.title('nn = {0} of {1}'.format(nn, n1))
    plt.show(block=False)
    plt.pause(1e-3)

